/*
  Abundance DHT

 Based on:
 https://www.electronicwings.com/avr-atmega/dht11-sensor-interfacing-with-atmega16-32

http://academy.cba.mit.edu/classes/embedded_programming/hello.ftdi.44.echo.c

http://academy.cba.mit.edu/classes/input_devices/button/hello.button.45.c

 modified 19 June 2021
 by Jean-Luc Pierite

*/
#include <avr/io.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <stdio.h>

#define serial_port PORTA
#define serial_direction DDRA
#define serial_pins PINA
#define serial_TX (1 << PA0)
#define serial_DHT (1 << PA3)

#define output(directions,pin) (directions |= pin) // set port direction for output
#define input(directions,pin) (directions &= (~pin)) // set port direction for input
#define set(port,pin) (port |= pin) // set port pin
#define clear(port,pin) (port &= (~pin)) // clear port pin
#define pin_test(pins,pin) (pins & pin) // test for port pin
#define bit_test(byte,bit) (byte & (1 << bit)) // test for bit set
#define bit_delay_time 102 // bit delay for 9600 with overhead
#define bit_delay() _delay_us(bit_delay_time) // RS232 bit delay
#define half_bit_delay() _delay_us(bit_delay_time/2) // RS232 half bit delay

uint8_t _data, high_Humid,low_Humid,high_Temp,low_Temp,checkSum; // global variables to store data

void put_char(volatile unsigned char *port, unsigned char pin, char txchar) {
   //
   // send character in txchar on port pin
   //    assumes line driver (inverts bits)
   //
   // start bit
   //
   clear(*port,pin);
   bit_delay();
   //
   // unrolled loop to write data bits
   //
   if bit_test(txchar,0)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,1)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,2)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,3)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,4)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,5)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,6)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   if bit_test(txchar,7)
      set(*port,pin);
   else
      clear(*port,pin);
   bit_delay();
   //
   // stop bit
   //
   set(*port,pin);
   bit_delay();
   //
   // char delay
   //
   bit_delay();
   }

void put_string(volatile unsigned char *port, unsigned char pin, char *str) {
   //
   // print a null-terminated string
   //
   static int index;
   index = 0;
   do {
      put_char(port, pin, str[index]);
      ++index;
      } while (str[index] != 0);
   }

void request_DHT() {
  output(serial_direction, serial_DHT);

  clear(serial_port, serial_DHT);	/* set to low pin */
	_delay_ms(20);			/* wait for 20ms */
  set(serial_port, serial_DHT);	/* set to high pin */
}

void response_DHT() {
  input(serial_direction, serial_DHT);

	while(serial_pins & serial_DHT); //wait for high pulse
	while((serial_pins & serial_DHT)==0); //wait for low pulse
	while(serial_pins & serial_DHT); //wait for high pulse
}

uint8_t read_Data() {
  for (int q=0; q<8; q++)
  	{
  		while((serial_pins & serial_DHT) == 0);  /* check received bit 0 or 1 */
  		_delay_us(30);
  		if(serial_pins & serial_DHT)/* if high pulse is greater than 30ms */
  		_data = (_data<<1)|(0x01);	/* then its logic HIGH */
  		else			/* otherwise its logic LOW */
  		_data = (_data<<1);
  		while(serial_pins & serial_DHT); // wait for high pulse
  	}
  return _data;
}

void start_test() {

  request_DHT();
  response_DHT();

  // read data into global variables
  high_Humid = read_Data();
  low_Humid = read_Data();
  high_Temp = read_Data();
  low_Temp = read_Data();
  checkSum = read_Data();
}

int main(void) {

  static char success[7] = "success";

  //
  // initialize TX pins
  //
  set(serial_port, serial_TX);
  output(serial_direction, serial_TX);
  set(serial_port, serial_DHT);
  _delay_ms(1000);

  //
  // set clock divider to /1
  //
  CLKPR = (1 << CLKPCE);
  CLKPR = (0 << CLKPS3) | (0 << CLKPS2) | (0 << CLKPS1) | (0 << CLKPS0);

  // main loop
  while (1) {
    start_test();

    if ((high_Humid + low_Humid + high_Temp + low_Temp) == checkSum)
    {
      put_string(&serial_port, serial_TX, success); // send success msg
      put_char(&serial_port, serial_TX, 0x0a); // new line
    }


    _delay_ms(5000);
  }

}
